module("L_EnOceanESP3", package.seeall)

local bit = require("nixio").bit

-------------------------------------------------------------------------------
-- Flags
-------------------------------------------------------------------------------

local DEBUG_MODE          = true
local WAIT_FOR_RESPONSE   = false -- If this is 'true', no further messages will be sent
local SENDING_IN_PROGRESS = false

-------------------------------------------------------------------------------
-- Constants
-------------------------------------------------------------------------------

-- Services
local SID = {
	ENOCEAN = "urn:micasaverde-com:serviceId:EnoceanGatewayESP3",
	SES     = "urn:micasaverde-com:serviceId:SecuritySensor1"
}

-- Message types
local SYS_MESSAGE_TYPES = {
	BUSY    =  1,
	ERROR   =  2,
	SUCCESS =  4
}

-- Delays for luup.delay, in seconds
local DELAYS = {
	PROCESS_PACKET    = 1,
	SEND_NEXT_MESSAGE = 2
}

-------------------------------------------------------------------------------

local SYNC_BYTE = string.char(0x55)

local STATE = {
	GET_SYNC_STATE    = 0,
	GET_HEADER_STATE  = 1,
	CHECK_CRC8H_STATE = 2,
	GET_DATA_STATE    = 3,
	CHECK_CRC8D_STATE = 4
}

local PACKET_TYPES = {
	RADIO          = string.char(0x1),
	RESPONSE       = string.char(0x2),
	EVENT          = string.char(0x4),
	COMMON_COMMAND = string.char(0x5)
}

local TELEGRAM_TYPES = {
	["RPS"] = string.char(0xF6),
	["1BS"] = string.char(0xD5),
	["4BS"] = string.char(0xA5)
}

local CRC8_TABLE = {
	0x00, 0x07, 0x0e, 0x09, 0x1c, 0x1b, 0x12, 0x15,
	0x38, 0x3f, 0x36, 0x31, 0x24, 0x23, 0x2a, 0x2d,
	0x70, 0x77, 0x7e, 0x79, 0x6c, 0x6b, 0x62, 0x65,
	0x48, 0x4f, 0x46, 0x41, 0x54, 0x53, 0x5a, 0x5d,
	0xe0, 0xe7, 0xee, 0xe9, 0xfc, 0xfb, 0xf2, 0xf5,
	0xd8, 0xdf, 0xd6, 0xd1, 0xc4, 0xc3, 0xca, 0xcd,
	0x90, 0x97, 0x9e, 0x99, 0x8c, 0x8b, 0x82, 0x85,
	0xa8, 0xaf, 0xa6, 0xa1, 0xb4, 0xb3, 0xba, 0xbd,
	0xc7, 0xc0, 0xc9, 0xce, 0xdb, 0xdc, 0xd5, 0xd2,
	0xff, 0xf8, 0xf1, 0xf6, 0xe3, 0xe4, 0xed, 0xea,
	0xb7, 0xb0, 0xb9, 0xbe, 0xab, 0xac, 0xa5, 0xa2,
	0x8f, 0x88, 0x81, 0x86, 0x93, 0x94, 0x9d, 0x9a,
	0x27, 0x20, 0x29, 0x2e, 0x3b, 0x3c, 0x35, 0x32,
	0x1f, 0x18, 0x11, 0x16, 0x03, 0x04, 0x0d, 0x0a,
	0x57, 0x50, 0x59, 0x5e, 0x4b, 0x4c, 0x45, 0x42,
	0x6f, 0x68, 0x61, 0x66, 0x73, 0x74, 0x7d, 0x7a,
	0x89, 0x8e, 0x87, 0x80, 0x95, 0x92, 0x9b, 0x9c,
	0xb1, 0xb6, 0xbf, 0xb8, 0xad, 0xaa, 0xa3, 0xa4,
	0xf9, 0xfe, 0xf7, 0xf0, 0xe5, 0xe2, 0xeb, 0xec,
	0xc1, 0xc6, 0xcf, 0xc8, 0xdd, 0xda, 0xd3, 0xd4,
	0x69, 0x6e, 0x67, 0x60, 0x75, 0x72, 0x7b, 0x7c,
	0x51, 0x56, 0x5f, 0x58, 0x4d, 0x4a, 0x43, 0x44,
	0x19, 0x1e, 0x17, 0x10, 0x05, 0x02, 0x0b, 0x0c,
	0x21, 0x26, 0x2f, 0x28, 0x3d, 0x3a, 0x33, 0x34,
	0x4e, 0x49, 0x40, 0x47, 0x52, 0x55, 0x5c, 0x5b,
	0x76, 0x71, 0x78, 0x7f, 0x6A, 0x6d, 0x64, 0x63,
	0x3e, 0x39, 0x30, 0x37, 0x22, 0x25, 0x2c, 0x2b,
	0x06, 0x01, 0x08, 0x0f, 0x1a, 0x1d, 0x14, 0x13,
	0xae, 0xa9, 0xa0, 0xa7, 0xb2, 0xb5, 0xbc, 0xbb,
	0x96, 0x91, 0x98, 0x9f, 0x8a, 0x8D, 0x84, 0x83,
	0xde, 0xd9, 0xd0, 0xd7, 0xc2, 0xc5, 0xcc, 0xcb,
	0xe6, 0xe1, 0xe8, 0xef, 0xfa, 0xfd, 0xf4, 0xf3
}


-------------------------------------------------------------------------------
-- Globals
-------------------------------------------------------------------------------

local lug_device = nil

local g_taskHandle = -1
local g_lastSysMessage = 0 -- Timestamp of the last status message.

local g_messageQueue = {}


-------------------------------------------------------------------------------
-- Generic utilities
-------------------------------------------------------------------------------

local log      = luup.log
local getinfo  = debug.getinfo -- Get the 'getinfo' function from the 'debug' module before overwriting the 'debug' keyword.
local function debug() end


local function showSysMessage (message, mode, critical)
	log("(EnOceanPlugin::showSysMessage/"..(getinfo(2).name or "N/A")..") message='".. message .."', mode=".. mode ..", critical=".. critical)

	luup.task(text, mode, "EnOcean Plugin", g_taskHandle)
	g_lastSysMessage = tostring(os.time())

	if (not critical) then
		luup.call_delay("clearSysMessage", 30, g_lastSysMessage) -- Clear the previous system message, since it's transient.
	else
		luup.set_failure(true, lug_device)
	end
end


function clearSysMessage (messageTime)
	-- 'messageTime' is nil if the function is called by the user.
	if (messageTime == g_lastSysMessage or messageTime == nil) then
		luup.task("Clearing...", SYS_MESSAGE_TYPES.SUCCESS, "EnOcean Plugin", g_taskHandle)
	end
end

--[[
local function padLeft (s, length, c)
	s = tostring(s)
	length = length or 2
	c = c or "0"
	return c:rep(length - #s) .. s
end


function table.contains (t, item)
	for k, v in pairs(t) do
		if (v == item) then
			return true, k
		end
	end

	return false
end

-- Set "class".
local set = {
	add = function (s, item)
	          if (not table.contains(s, item)) then
	              s[#s + 1] = item
	          end
	      end,
	remove = function (s, item)
	             local item, pos = table.contains(t, item)
	             if (item) then
	                 table.remove(t, pos)
	             end
	         end
}
]]

-------------------------------------------------------------------------------
-- Plugin specific functions
-------------------------------------------------------------------------------

local function computeCrc (data, start, length)
	start = start or 1
	length = length or #data

	local crc = 0
	for i = start, start + length - 1 do
		crc = CRC8_TABLE[bit.bxor(crc, string.byte(data[i])) + 1]
	end
	debug("(EnOceanPlugin::computeCrc) crc=".. crc)
	return string.char(crc)
end


local function addToSendQueue (packetType, data, optData, waitForResponse)
	local insert = table.insert
	local char   = string.char

	local packet = { SYNC_BYTE }             -- The sync byte

	insert(packet, char(#data / 0x100))      -- First byte of the data length
	insert(packet, char(#data % 0x100))      -- Seconds byte of the data length
	insert(packet, char(#optData))           -- Length of the optional data
	insert(packet, packetType)               -- The packet type
	insert(packet, computeCrc(packet, 2, 4)) -- The header CRC8

	insert(packet, data)                     -- The user data
	insert(packet, optData)                  -- The optional user data
	insert(packet, computeCrc(packet, 7, #data + #optData)) -- The data CRC8

	table.insert(g_messageQueue, {content = table.concat(packet), waitForResponse = (waitForResponse or false)})
	if (not SENDING_IN_PROGRESS) then
		sendPacket()
	end
end


function sendPacket()
	SENDING_IN_PROGRESS = true

	if (WAIT_FOR_RESPONSE == true) then
		g_triesCount = g_triesCount + 1
		if(g_triesCount < 3) then
			debug("(EnOceanPlugin::sendPacket) WAIT_FOR_RESPONSE flag is set, g_triesCount=".. g_triesCount .."; try again in ".. DELAYS.SEND_NEXT_MESSAGE .." seconds.")
			luup.call_delay("sendPacket", DELAYS.SEND_NEXT_MESSAGE)
			return
		else
			log("(EnOceanPlugin::sendPacket) Didn't receive response for the last message. Stop waiting for response.")
			WAIT_FOR_RESPONSE = false
			g_triesCount = 0
		end
	end

	local message
	if (#g_messageQueue > 0) then
		message = g_messageQueue[1]
	else
		return
	end

	debug("(EnOceanPlugin::sendPacket) Wait for response: ".. message.waitForResponse)

	if (not luup.io.write(message.content)) then
		showSysMessage("Failed to send message", SYS_MESSAGE_TYPES.ERROR)
		luup.variable_set(SID.ENOCEAN, "CommFailure", "1", lug_device)
		return
	end

	WAIT_FOR_RESPONSE = message.waitForResponse
	table.remove(g_messageQueue, 1)

	-- Send the next message in the queue.
	if (#g_messageQueue > 0) then
		luup.call_delay("sendPacket", DELAYS.SEND_NEXT_MESSAGE)
	else
		SENDING_IN_PROGRESS = false
	end
end


-------------------------------------------------------------------------------
-- Radio telegram handlers
-------------------------------------------------------------------------------

local RADIO_TELEGRAM_HANDLERS = {
	-- RPS
	[TELEGRAM_TYPES.RPS] = function (data)
		debug("(EnOceanPlugin::RPS) RPS telegram handler")
	end,

	-- 1BS
	[TELEGRAM_TYPES["1BS"]] = function (data)
		debug("(EnOceanPlugin::1BS) 1BS telegram handler")

	end,

	-- 4BS
	[TELEGRAM_TYPES.4BS] = function (data)
		debug("(EnOceanPlugin::4BS) 4BS telegram handler")
	end
}


-------------------------------------------------------------------------------
-- Packet handler functions
-------------------------------------------------------------------------------

function handleRadioTelegram (data)
	debug("(EnOceanPlugin::handleRadioTelegram) Handle radio telegram")
	RADIO_TELEGRAM_HANDLERS[string.byte(data:sub(1, 1))](data:sub(2))
end


function handleResponse (data)
	debug("(EnOceanPlugin::handleResponse) Handle response")
end


local PACKET_HANDLERS = {
	[PACKET_TYPES.RADIO]    = "handleRadioTelegram",
	[PACKET_TYPES.RESPONSE] = "handleResponse"
}


-------------------------------------------------------------------------------
-- Handle incoming data
-------------------------------------------------------------------------------

-- 4BS
-- 0x55 | 0x0 0xa 0x7 0x1 | 0xeb | 0xa5 - 0xce 0x79 0x0 0xf - 0x0 0x81 0x1 0x75 - 0x0 | 0x1 0xff 0xff 0xff 0xff 0x44 0x0 | 0x6e
-- 0x55 | 0x0 0xa 0x7 0x1 | 0xeb | 0xa5 - 0xc6 0x79 0x0 0xf - 0x0 0x81 0x1 0x75 - 0x0 | 0x1 0xff 0xff 0xff 0xff 0x43 0x0 | 0x15
-- 0x55 | 0x0 0xa 0x7 0x1 | 0xeb | 0xa5 - 0xc6 0x79 0x0 0xf - 0x0 0x81 0x1 0x75 - 0x0 | 0x1 0xff 0xff 0xff 0xff 0x43 0x0 | 0x15

-- 1BS
-- 0x55 | 0x0 0x7 0x7 0x1 | 0x7a | 0xd5 - 0x8 - 0x0 0x1 0x8a 0xbb - 0x0 | 0x1 0xff 0xff 0xff 0xff 0x3a 0x0 | 0xbc
-- 0x55 | 0x0 0x7 0x7 0x1 | 0x7a | 0xd5 - 0x8 - 0x0 0x0 0x85 0xe - 0x0 | 0x1 0xff 0xff 0xff 0xff 0x37 0x0 | 0x3a


local state        = STATE.GET_SYNC_STATE -- The current state
local rxBuf        = {} -- The received data buffer
local rxCount      = 0  -- The number of bytes in the data buffer
local rxDataLen    = 0  -- The length of the data field
local rxOptDataLen = 0  -- The length of the optional data field
local rxPacketType = 0  -- The packet type

-- State machine to load incoming packet bytes
local STATE_MACHINE = {

	-- Waiting for packet sync byte 0x55
	[STATE.GET_SYNC_STATE] = function (rxByte)
		if (rxByte == SYNC_BYTE) then
			debug("(EnOceanPlugin::GET_SYNC_STATE) Got sync byte, go to GET_HEADER_STATE")
			state = STATE.GET_HEADER_STATE
			crc   = 0
			rxCount = 0
		end
	end,

	-- Read the header bytes.
	[STATE.GET_HEADER_STATE] = function (rxByte)
		-- Add the received byte in the buffer.
		rxCount = rxCount + 1
		rxBuf[rxCount] = rxByte
		-- All the header bytes received?
		if(rxCount == 4) then
			debug("(EnOceanPlugin::GET_HEADER_STATE) Received all header bytes, go to CHECK_CRC8H_STATE")
			state = STATE.CHECK_CRC8H_STATE
		end
	end,

	-- Check the header checksum & try to resynchronize if and error occured.
	[STATE.CHECK_CRC8H_STATE] = function (rxByte)
		-- Header CRC correct?
        if (computeCrc(rxBuf, 1, rxCount) ~= rxByte) then
			debug("(EnOceanPlugin::CHECK_CRC8H_STATE) Incorrect header crc")
			-- No. Check if there is a sync byte (0x55) in the header.
			local pos = -1
			for i = 1, 4 do
				if (rxBuf[i] == SYNC_BYTE) then
					-- Indicates the position of the sync byte
					pos = i
					break
				end
			end

			if (pos == -1 and rxByte ~= SYNC_BYTE) then
				-- Neither the header nor the CRC8H contain the sync byte
				debug("(EnOceanPlugin::CHECK_CRC8H_STATE) Neither the header nor the CRC8H contain the sync byte, go to GET_SYNC_STATE")
				state = STATE.GET_SYNC_STATE
				return
			elseif (pos == -1 and rxByte == SYNC_BYTE) then
				-- The header does not contain the sync byte but the CRC8H does
				-- The sync byte could be the beginning of a packet
				debug("(EnOceanPlugin::CHECK_CRC8H_STATE) The header does not contain the sync byte but the CRC8H does, go to GET_HEADER_STATE")
				state = STATE.GET_HEADER_STATE
				rxCount = 0
				return
			end

			-- The header contains the sync byte. It could be a new packed.
			-- Shift all the bytes from the 0x55 byte in the buffer.
			-- Recalculate CRC8 for those bytes.
			for i = 1, 4 - pos do
				rxBuf[i] = rxBuf[pos + i]
			end
			rxCount = 4 - pos

			-- Copy the received byte to buffer
			rxCount = rxCount + 1
			rxBuf[rxCount] = rxByte

			if (rxCount < 4) then
				debug("(EnOceanPlugin::CHECK_CRC8H_STATE) There are still header bytes to receive, go to GET_HEADER_STATE")
				state = STATE.GET_HEADER_STATE
			end

			debug("(EnOceanPlugin::CHECK_CRC8H_STATE) All header bytes received, stay in CHECK_CRC8H_STATE")
			return
        end

		-- CRC8H correct. Length fields values valid?
		rxDataLen = string.byte(rxBuf[1]) * 256 + string.byte(rxBuf[2])
		rxOptDataLen = string.byte(rxBuf[3])
		debug("(EnOceanPlugin::CHECK_CRC8H_STATE) Header crc correct, rxDataLen=".. rxDataLen ..", rxOptDataLen=".. rxOptDataLen)
		if ((rxDataLen + rxOptDataLen) == 0) then
			-- No. Sync byte received?
			if (rxByte == SYNC_BYTE) then
				-- Yes
				debug("(EnOceanPlugin::CHECK_CRC8H_STATE) Header crc incorrect, sync byte not received, go to GET_HEADER_STATE")
				state = STATE.GET_HEADER_STATE
				rxCount = 0
				return
			end

			-- Packet with correct CRC8H, but wrong length fields
			debug("(EnOceanPlugin::CHECK_CRC8H_STATE) Packet with correct CRC8H, but wrong length fields, go to GET_SYNC_STATE")
			state = STATE.GET_SYNC_STATE
			debug("(EnOceanPlugin::CHECK_CRC8H_STATE) Error: invalid packet")
			return
		end

		-- Correct header CRC8. Go to the reception of data.
		debug("(EnOceanPlugin::CHECK_CRC8H_STATE) Correct header CRC8, go to GET_DATA_STATE")
		state = STATE.GET_DATA_STATE
		rxCount = 0
		rxPacketType = string.byte(rxBuf[4])
	end,

	-- Copy the information bytes.
	[STATE.GET_DATA_STATE] = function (rxByte)
		rxCount = rxCount + 1
		rxBuf[rxCount] = rxByte

		-- When all the expected bytes are received, go to calculate data checksum.
		if (rxCount == (rxDataLen + rxOptDataLen)) then
			debug("(EnOceanPlugin::GET_DATA_STATE) All expected bytes received, go to CHECK_CRC8D_STATE")
			state = STATE.CHECK_CRC8D_STATE
		end
	end,

	-- Check the data CRC8
	[STATE.CHECK_CRC8D_STATE] = function (rxByte)
		-- In all cases the state returns to the first state: waiting for next sync byte.
		state = STATE.GET_SYNC_STATE

		if (computeCrc(rxBuf, 1, rxCount) == rxByte) then
			-- Correct packet received
			debug("(EnOceanPlugin::CHECK_CRC8D_STATE) Correct packet received, schedule the packet handler function and go to GET_SYNC_STATE")
			if (PACKET_HANDLERS[rxPacketType] ~= nil) then
				luup.call_delay(PACKET_HANDLERS[rxPacketType], DELAYS.PROCESS_PACKET, table.concat(rxBuf, "", 1, rxCount))
			else
				debug("(EnOceanPlugin::CHECK_CRC8D_STATE) Unhandled packet type ".. rxPacketType)
			end
			return
		end

		-- Incorrect CRC8
		-- If the received byte is the sync byte, then it could be the sync byte for next packet.
		if (rxByte == SYNC_BYTE) then
			debug("(EnOceanPlugin::CHECK_CRC8D_STATE) Incorrect CRC8, the received byte is the sync byte, go to GET_HEADER_STATE")
			state = STATE.GET_HEADER_STATE
			rxCount = 0
		end

		log("(EnOceanPlugin::CHECK_CRC8D_STATE) Incorrect CRC8")
		log("(EnOceanPlugin::CHECK_CRC8D_STATE) Error: incorrect data checksum")
	end
}


function handleIncoming (lul_data)
	log("state=".. state ..", rxByte=".. (string.byte(lul_data) or "nil"))
	if (STATE_MACHINE[state] ~= nil) then
		STATE_MACHINE[state](lul_data)
	else
		-- Wait for the sync byte.
		state = STATE.GET_SYNC_STATE
	end
end


-------------------------------------------------------------------------------
-- Action implementations
-------------------------------------------------------------------------------

function setTarget (device, newTargetValue)
	--local
end


-----------------------------------------------------------------------------
-- Init
-------------------------------------------------------------------------------

function init (lul_device)
	log("(EnOceanPlugin::init) EnOcean ESP3 Gateway plugin")

	-- Get the root device.
	lug_device = lul_device

	-- Get a handle for system messages.
	g_taskHandle = luup.task("Starting up...", 1, "EnOcean plugin", -1)

	-- Reset our CommFailure variable.
	luup.variable_set(SID.ENOCEAN, "CommFailure", "0", lul_device)

	-- Update the Debug Mode.
	local debugMode = luup.variable_get(ENOCEAN_SID, "DebugMode", lul_device) or ""
	if (debugMode == "") then
		luup.variable_set(ENOCEAN_SID, "DebugMode", (DEBUG_MODE and "1" or "0"), lul_device)
	else
		DEBUG_MODE = (debugMode == "1") and true or false
	end

	if (DEBUG_MODE) then
		debug = log
	end

	-- Get the sender ID of the TCM300.
	addToSendQueue(PACKET_TYPES.COMMON_COMMAND, COMMON_COMMAND.CO_RD_IDBASE, "", true)

	debug("(EnOceanPlugin::init) Success: startup successful")
	return true, "startup successful", "EnOcean plugin"
end
